/* soundmosaic, Copyright (c) 2001-2007 Steven Hazel <sah@thalassocracy.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */
#include <stdio.h>
#include <stdint.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "sm_utils.h"


int interlace (const int16_t *in, int16_t *out, int length, int channels)
{
  int frames, i, j;

  assert(channels > 0);

  if (channels == 1) {
    memcpy(out, in, length * sizeof(int16_t));
    return 1;
  }

  frames = length/channels;

  for (i=0; i<frames; i++) {
    for (j=0; j<channels; j++) {
      out[(i*channels)+j] = in[(j*frames)+i];
    }
  }

  return 1;
}


int deinterlace (const int16_t *in, int16_t *out, int length, int channels)
{
  int frames, i, j;

  assert(channels > 0);

  if (channels == 1) {
    memcpy(out, in, length * sizeof(int16_t));
    return 1;
  }

  frames = length/channels;

  for (i=0; i<frames; i++) {
    for (j=0; j<channels; j++) {
      out[(j*frames)+i] = in[(i*channels)+j];
    }
  }

  return 1;
}


void range(const int16_t *data, int32_t length, int16_t *max, int16_t *min)
{
  int32_t i;

  assert(length > 0);

  *min = INT16_MAX;
  *max = INT16_MIN;
  for (i = 0; i < length; i++) {
    if (data[i] > *max) {
      *max = data[i];
    }
    if (data[i] < *min) {
      *min = data[i];
    }
  }

  assert(*max >= *min);
}


/* amplitude is defined as the size of the range.  we need an int32_t
   to represent the maximum range, because INT16_MAX - INT16_MIN =
   UINT16_MAX + 1. (so, if the range were -1 to 1, the maxmum
   amplitude would be 2...) */
int32_t amplitude_of_range(int16_t max, int16_t min)
{
  int32_t retval;

  retval = ((int32_t)max - (int32_t)min);

  /* we should never return an odd amplitude */
  if (retval % 2 != 0) {
    retval++;
  }

  assert(retval >= 0);
  assert(retval <= UINT16_MAX + 1);

  return retval;
}


int32_t amplitude(const int16_t *data, int32_t length)
{
  int16_t max, min;

  assert(length > 0);

  range(data, length, &max, &min);

  return amplitude_of_range(max, min);
}


int16_t middle(const int16_t *data, int32_t length)
{
  int16_t max, min;

  assert(length > 0);

  range(data, length, &max, &min);

  return min + ((max - min) / 2);
}



int16_t mean(const int16_t *data, int32_t length)
{
  int32_t i;
  int16_t avg;
  int32_t rem;

  rem = 0;
  avg = 0;
  for (i = 0; i < length; i++) {
    avg += data[i] / length;
    rem += data[i] % length;
    if (abs(rem) >= length) {
      avg += rem / length;
      rem = rem % length;
    }
  }

  assert(rem < length);

  return avg;
}



void mean_shift(const int16_t *in, int16_t *out, int32_t length,
                int16_t newmean)
{
  int32_t i;
  int16_t origmean;
  int32_t distance;
  int16_t max, min;

  assert(length > 0);

  range(in, length, &max, &min);

  origmean = mean(in, length);

  distance = newmean - origmean;

  /* Don't let distance be greater than the distance between the max
     and absolute max or min and absolute min, as appropriate to the
     direction of distance */
  if (distance > (INT16_MAX - max)) {
    assert(distance > 0);
    distance = INT16_MAX - max;
  } else if ((-distance) > (min - INT16_MIN)) {
    assert(distance < 0);
    distance = INT16_MIN - min;
    assert(distance == -(min - INT16_MIN));
  }

  newmean = origmean + distance;

  for (i = 0; i < length; i++) {
    out[i] = in[i] + distance;
  }

  //printf("%d + %d = %d, %d\n", origmean, distance, newmean, mean(out, length));
  assert(mean(out, length) >= newmean - 1);
  assert(mean(out, length) <= newmean + 1);
}


void scale_ampl(const int16_t *in, int16_t *out, int32_t length,
                int32_t newampl)
{
  int32_t i;
  int32_t origampl;
  int16_t max, min;
  int16_t tempbuffer[length];
  /* int32_t testedampl; */
  double val;

  assert(length > 0);

  mean_shift(in, tempbuffer, length, 0);

  range(tempbuffer, length, &max, &min);
  origampl = amplitude_of_range(max, min);

  for (i = 0; i < length; i++) {
    val = (double)tempbuffer[i];
    val /= origampl;
    val *= newampl;
    out[i] = (int16_t)val;

    /* ensure that the scaling isn't changing the sign of the data
      (since that should never happen...).  We can't assert that the
      result isn't zero, since any value might be low enough to get
      scaled to 0. */
    if (val > 0) {
      assert(tempbuffer[i] >= 0);
      assert(out[i] >= 0);
    } else if (val < 0 ) {
      assert(tempbuffer[i] <= 0);
      assert(out[i] <= 0);
    }
  }

  /* testedampl = amplitude(out, length);
     printf("started with %d, expected %d, got %d\n", origampl, newampl,
            testedampl);
     assert(testedampl == newampl);
  */
}

int64_t correlation(const int16_t *a, const int16_t *b, int32_t length)
{
  int64_t corr, oldcorr, prod;
  int32_t i;

  /* Because length is an int32_t, it's impossible for corr to
     overflow.  corr is the sum of length products of two int16_ts.
     corr is stored as an int64_t.  The worst case is that every
     product is INT16_MIN squared, and there are INT32_MAX of them --
     that's less than INT32_MAX squared, which is less than
     INT64_MAX. */

  corr = 0;
  for (i = 0; i < length; i++) {
    prod = a[i] * b[i];
    oldcorr = corr;
    corr += prod;
    if (prod < 0) {
      assert(corr < oldcorr);
    } else if (prod == 0) {
      assert(oldcorr == corr);
    } else if (prod > 0) {
      assert(corr > oldcorr);
    }
  }

  return corr;
}


int64_t magnitude(const int16_t *data, int32_t length)
{
  int64_t corr;
  double mag;

  /* correlation is just a dot product, so autocorrelation is
     magnitude squared */
  corr = correlation(data, data, length);
  mag = sqrt((double)corr);
  assert(mag >= 0);

  return (int64_t)mag;
}


void scale_mag(const int16_t *in, int16_t *out, int32_t length,
               int64_t newmag)
{
  int32_t i;
  int64_t origmag;
  double val;

  assert(length > 0);
  assert(newmag >= 0);

  origmag = magnitude(in, length);

  for (i = 0; i < length; i++) {
    val = (double)in[i];
    val /= origmag;
    val *= newmag;

    if (val > INT16_MAX) {
      val = INT16_MAX;
    } else if (val < INT16_MIN) {
      val = INT16_MIN;
    }

    out[i] = (int16_t)val;

    /* ensure that the scaling isn't changing the sign of the data
      (since that should never happen...).  We can't assert that the
      result isn't zero, since any value might be low enough to get
      scaled to 0. */
    if (val > 0) {
      assert(in[i] >= 0);
      assert(out[i] >= 0);
    } else if (val < 0) {
      assert(in[i] <= 0);
      assert(out[i] <= 0);
    }
  }
}
